from pydantic import BaseModel


class Signs(BaseModel):
    debit: int
    credit: int


def signs(ledger_type: int) -> Signs:
    if ledger_type == 0:  # ASSETS
        return Signs(debit=1, credit=-1)
    elif ledger_type == 1:  # CAPITAL
        return Signs(debit=-1, credit=1)
    elif ledger_type == 2:  # LIABILITIES
        return Signs(debit=-1, credit=1)
    elif ledger_type == 3:  # REVENUE
        return Signs(debit=-1, credit=1)
    elif ledger_type == 4:  # EXPENSES
        return Signs(debit=1, credit=-1)
    else:
        raise RuntimeError(f"Unhandled ledger_type: {ledger_type}")



from django.db import models
from django.conf import settings
from django.utils.translation import gettext as _


class Ledger(models.Model):
    LEDGER_TYPE_ASSETS = 0
    LEDGER_TYPE_CAPITAL = 1
    LEDGER_TYPE_LIABILITIES = 2
    LEDGER_TYPE_REVENUE = 3
    LEDGER_TYPE_EXPENSES = 4

    LEDGER_TYPES = [
        (LEDGER_TYPE_ASSETS, _("Assets")),
        (LEDGER_TYPE_CAPITAL, _("Capital")),
        (LEDGER_TYPE_LIABILITIES, _("Liabilities")),
        (LEDGER_TYPE_REVENUE, _("Revenue")),
        (LEDGER_TYPE_EXPENSES, _("Expenses"))
    ]

    ledger_type = models.IntegerField(choices=LEDGER_TYPES)

    name = models.CharField(
        help_text=_("Name of this ledger"),
        max_length=256
    )

    description = models.TextField(
        help_text=_("Any notes to go along with this Transaction."),
        blank=True
    )

    def __str__(self):
        return f"{self.name}"


class Transaction(models.Model):
    TRANSACTION_MANUAL = 0
    TRANSACTION_AUTOMATED = 1

    TRANSACTION_TYPES = [
        (TRANSACTION_MANUAL, _("Manual")),
        (TRANSACTION_AUTOMATED, _("Automated")),
    ]

    transaction_type = models.IntegerField(choices=TRANSACTION_TYPES)

    notes = models.TextField(
        help_text=_("Any notes to go along with this Transaction."),
        blank=True
    )
    voids = models.OneToOneField(
        "Transaction",
        blank=True,
        null=True,
        related_name="voided_by",
        on_delete=models.deletion.CASCADE
    )

    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.deletion.CASCADE
    )

    posted_timestamp = models.DateTimeField(
        help_text=_("Time the transaction was posted. Change this field to model retroactive ledger entries."),
        db_index=True
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(auto_now=True, blank=True, null=True)

    @property
    def balanced(self):
        entries = self.entries.all()
        credits = sum([e.credit if e.credit else 0 for e in entries])
        debits = sum([e.debit if e.debit else 0 for e in entries])
        return credits == debits

    def __str__(self):
        return f"ID: {self.id}"


class TransactionEvidence(models.Model):
    OBJECT_TYPE_ORDER = 0

    OBJECT_TYPES = [
        (OBJECT_TYPE_ORDER, _("Order"))
    ]

    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE, related_name="evidences")
    object_type = models.IntegerField(choices=OBJECT_TYPES)
    object_id = models.IntegerField()

    def __str__(self):
        return f"Type: {self.object_type}, Object ID: {self.object_id} to Transaction: {self.transaction}"


class LedgerEntry(models.Model):
    ledger = models.ForeignKey(Ledger, on_delete=models.CASCADE, related_name="entries")
    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE, related_name="entries")

    debit = models.DecimalField(max_digits=15, decimal_places=2, null=True, blank=True)
    credit = models.DecimalField(max_digits=15, decimal_places=2, null=True, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True, blank=True, null=True)

    def __str__(self):
        return f"Entry: DEBIT: {self.debit} CREDIT: {self.credit} Ledger: {self.ledger}"



# Generated by Django 4.2 on 2023-06-30 10:40

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Ledger',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ledger_type', models.IntegerField(choices=[(0, 'Assets'), (1, 'Capital'), (2, 'Liabilities'), (3, 'Revenue'), (4, 'Expenses')])),
                ('name', models.CharField(help_text='Name of this ledger', max_length=256)),
                ('description', models.TextField(blank=True, help_text='Any notes to go along with this Transaction.')),
            ],
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('transaction_type', models.IntegerField(choices=[(0, 'Manual'), (1, 'Automated')])),
                ('notes', models.TextField(blank=True, help_text='Any notes to go along with this Transaction.')),
                ('posted_timestamp', models.DateTimeField(db_index=True, help_text='Time the transaction was posted. Change this field to model retroactive ledger entries.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('voids', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='voided_by', to='simple_accounting.transaction')),
            ],
        ),
        migrations.CreateModel(
            name='TransactionEvidence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_type', models.IntegerField(choices=[(0, 'Order')])),
                ('object_id', models.IntegerField()),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='evidences', to='simple_accounting.transaction')),
            ],
        ),
        migrations.CreateModel(
            name='LedgerEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('debit', models.DecimalField(blank=True, decimal_places=2, max_digits=15, null=True)),
                ('credit', models.DecimalField(blank=True, decimal_places=2, max_digits=15, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
                ('ledger', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='simple_accounting.ledger')),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='simple_accounting.transaction')),
            ],
        ),
    ]



from django.contrib import admin

from simple_accounting.models import *


class LedgerAdmin(admin.ModelAdmin):
    list_display = ("name", "ledger_type")


class TransactionAdmin(admin.ModelAdmin):
    list_display = ("transaction_type", "voids", "owner", "balanced", "posted_timestamp", "notes")


class LedgerEntryAdmin(admin.ModelAdmin):
    list_display = ("ledger", "transaction", "debit", "credit", "created_at")


admin.site.register(Ledger, LedgerAdmin)
admin.site.register(LedgerEntry, LedgerEntryAdmin)
admin.site.register(Transaction, TransactionAdmin)
admin.site.register(TransactionEvidence)
